#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"column.h"
#include	"transfrm.h"

/* ---------------------------------------------------------------------
	Initialise the transformation to do an empty (null) translation.
   --------------------------------------------------------------------- */

void	TransformInitialise(TRANSFORM* pTransform, int iRecLen, int iNewRecLen)
{
	int	i;

	pTransform->iNewRecLen = iNewRecLen;

	/* allocate memory to keep track of original and new column sequence */

	if ((pTransform->piOriginal = (int*) malloc(iRecLen * sizeof(int))) == 0)
		error_handler(4, "first permutation table");

	for (i=0; i<iRecLen; i++) 
		pTransform->piOriginal[i] = i+1;

	if ((pTransform->piPermute = (int*) malloc(iNewRecLen * sizeof(int))) == 0)
		error_handler(4, "second permutation table");

	for (i=0; i<iNewRecLen; i++) 
		pTransform->piPermute[i] = 0;

	/*	allocate memory to allow modification of selected columns */

	if ((pTransform->pchSet = (char*) malloc(iNewRecLen)) == 0)
		error_handler(4, "columns to set");

	for (i=0; i<iNewRecLen; i++) 
		pTransform->pchSet[i] = 0;

	if ((pTransform->pchValues = (char*) malloc(iNewRecLen)) == 0)
		error_handler(4, "new column values");

	for (i=0; i<iNewRecLen; i++) 
		pTransform->pchValues[i] = 0;
}

/* ---------------------------------------------------------------------
	Read the command line arguments to determine how to transform each
	column within the record.
   --------------------------------------------------------------------- */

void	TransformSetup(TRANSFORM* pTransform, int iRecLen, int argc, char** argv)
{
	int	i, j;
	int	*piDest, *piSource;
	int	cStart, cEnd, cNew;
	char	*parm;
	char	chType;

	for (i=3; i<argc; i++) {
		if (toupper(*argv[i]) == 'R') {
			parm		= argv[i]+1;
			cStart	= read_param(&parm, 1);
			cEnd		= read_param(&parm, iRecLen);
			chType	= *parm++;

			switch (toupper(chType)) {
				case 'C'	:
					if (cStart < 1 || cStart > iRecLen)
						error_handler(9,argv[i]);

					if (cEnd < 1 || cEnd > iRecLen)
						error_handler(9,argv[i]);

					cNew = read_param(&parm, 1);
					if (cNew < 1 || cNew > pTransform->iNewRecLen)
						error_handler(9,argv[i]);

					piDest = pTransform->piPermute + cNew-1;
					for (j=cStart; j<=cEnd; j++) {
						if (*piDest != 0) error_handler(8,argv[i]);
						*piDest++ = j;
					}
					break;

				case 'M' :
					if (cStart < 1 || cStart > iRecLen)
						error_handler(9,argv[i]);

					if (cEnd < 1 || cEnd > iRecLen)
						error_handler(9,argv[i]);

					cNew = read_param(&parm, 1);
					if (cNew < 1 || cNew > pTransform->iNewRecLen)
						error_handler(9,argv[i]);

					piDest	= pTransform->piPermute		+ cNew	- 1;
					piSource	= pTransform->piOriginal	+ cStart	- 1;

					for (j=cStart; j<=cEnd; j++) {
						if (*piDest != 0) error_handler(8,argv[i]);
						*piDest++	= j;
						*piSource++	= 0;
					}
					break;

				case 'D' :
					if (cStart < 1 || cStart > iRecLen)
						error_handler(9,argv[i]);

					if (cEnd < 1 || cEnd > iRecLen)
						error_handler(9,argv[i]);

					piSource = pTransform->piOriginal + cStart-1;
					for (j=cStart; j<=cEnd; j++) {
						if (*piSource == 0) error_handler(8,argv[i]);
						*piSource++ = 0;
					}
					break;

				case 'Z' :
					if (cStart < 1 || cStart > pTransform->iNewRecLen)
						error_handler(12,argv[i]);

					if (cEnd < 1 || cEnd > pTransform->iNewRecLen)
						error_handler(12,argv[i]);

					for (j=cStart-1; j<cEnd; j++) {
						pTransform->pchSet[j]		= 1;
						pTransform->pchValues[j]	= 0;
					}
					break;

				case 'B' :
					if (cStart < 1 || cStart > pTransform->iNewRecLen)
						error_handler(12,argv[i]);

					if (cEnd < 1 || cEnd > pTransform->iNewRecLen)
						error_handler(12,argv[i]);

					for (j=cStart-1; j<cEnd; j++) {
						pTransform->pchSet[j]		= 1;
						pTransform->pchValues[j]	= ' ';
					}
					break;

				case 'S' :
					if (cStart < 1 || cStart > pTransform->iNewRecLen)
						error_handler(12,argv[i]);

					if (cEnd < 1 || cEnd > pTransform->iNewRecLen)
						error_handler(12,argv[i]);

					for (j=cStart-1; j<cEnd; j++) {
						pTransform->pchSet[j]		= 1;
						pTransform->pchValues[j]	= (char) read_param(&parm, 0);
					}
					break;

				default	:
					error_handler(5,argv[i]);
					break;
			}
		}
	}

	piSource	= pTransform->piOriginal;
	piDest	= pTransform->piPermute - 1;

	for (i=0; i<iRecLen; i++) {
		if (*piSource != 0) {
			while (*++piDest);
			*piDest = *piSource;
		}
		piSource++;
	}
}

/* ---------------------------------------------------------------------
	Re-arrange the columns of a record.
   --------------------------------------------------------------------- */

void	translate(TRANSFORM* pTransform, const char *pchInput, char *pchOutput, BOOL fTextMode)
{
	int	i, j;
	char	ch;

	for (i=0; i<pTransform->iNewRecLen; i++) {
		j = pTransform->piPermute[i];
		ch = (j == 0) ? 0 : pchInput[j-1];
		if ((ch == 0) && fTextMode) break;
		*pchOutput++ = ch;
	}
	if (fTextMode) *pchOutput = 0;
}

/* ---------------------------------------------------------------------
	Replace selected columns of a record.
   --------------------------------------------------------------------- */

void	replace(TRANSFORM* pTransform, char *pchOutput)
{
	int		i;

	for (i=0; i<pTransform->iNewRecLen; i++) {
		if (pTransform->pchSet[i])
			pchOutput[i] = pTransform->pchValues[i];
	}
}

/* ---------------------------------------------------------------------
	Perform the complete transformation of a record.
   --------------------------------------------------------------------- */

void	TransformRecord(TRANSFORM* pTransform, const char *pchInput, char *pchOutput, BOOL fTextMode)
{
	memset(pchOutput, 0, pTransform->iNewRecLen);
	translate(pTransform, pchInput, pchOutput, fTextMode);
	replace(pTransform, pchOutput);
}

/* ---------------------------------------------------------------------
	Free any dynamically allocated structures.
   --------------------------------------------------------------------- */

void	TransformCleanup(TRANSFORM* pTransform)
{
	free(pTransform->piOriginal);
	free(pTransform->piPermute);
	free(pTransform->pchSet);
	free(pTransform->pchValues);
}
